from typing import Any, List, Mapping, MutableMapping, MutableSequence, Sequence, Union

import tinytim.data as data_features
import tinytim.copy as copy


def edit_row_items_inplace(data: MutableMapping, index: int, items: Mapping) -> None:
    """Changes row index to mapping items values.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_row_items_inplace(data, 0, {'x': 11, 'y': 66}) -> None
       data -> {'x': [11, 2, 3], 'y': [66, 7, 8]}
    """
    for col in items:
        data[col][index] = items[col]


def edit_row_values_inplace(data: MutableMapping, index: int, values: Sequence) -> None:
    """Changed row index to values.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_row_values_inplace(data, 1, (22, 77)) -> None
       data -> {'x': [1, 22, 3], 'y': [6, 77, 8]}
    """
    if len(values) != data_features.column_count(data):
        raise AttributeError('values length must match columns length.')
    for col, value in zip(data_features.column_names(data), values):
        data[col][index] = value


def edit_column_inplace(data: MutableMapping, column_name: str, values: MutableSequence) -> None:
    """Add values to data in named column.
       Overrides existing values if column exists,
       Created new column with values if column does not exist.

       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_column_inplace(data, 'x', [11, 22, 33]) -> None
       data -> {'x': [11, 22, 33], 'y': [6, 7, 8]}
    """
    if len(values) != data_features.row_count(data):
        raise ValueError('values length must match data rows count.')
    data[column_name] = values


def drop_row_inplace(data: MutableMapping, index: int) -> None:
    """Remove index row from data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       drop_row_inplace(data, 1) -> None
       data -> {'x': [1, 3], 'y': [6, 8]}
    """
    for col in data_features.column_names(data):
        data[col].pop(index)


def drop_label_inplace(labels: Union[None, List], index) -> None:
    """If labels exists, drop item at index.

       Examples:
       labels = [1, 2, 3, 4, 5]
       drop_label_inplace(labels, 1) -> None
       labels -> [1, 3, 4, 5]

       labels = None
       drop_label_inplace(labels, 1) -> None
       labels -> None
    """
    if labels is not None:
        labels.pop(index)


def drop_column_inplace(data: MutableMapping, column_name: str) -> None:
    """Return a new dict with the named column removed from data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       drop_column_inplace(data, 'y') -> None
       data -> {'x': [1, 2, 3]}
    """
    del data[column_name]


def edit_value_inplace(data: MutableMapping, column_name: str, index: int, value: Any) -> None:
    """Edit the value in named column as row index.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_value_inplace(data, 'x', 0, 11) -> None
       data -> {'x': [11, 2, 3], 'y': [6, 7, 8]}
    """
    data[column_name][index] = value


def replace_column_names(data: MutableMapping, new_names: Sequence[str]) -> dict:
    """Return a new dict same column data but new column names.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       replace_column_names(data, ('xx', 'yy')) -> {'xx': [1, 2, 3], 'yy': [6, 7, 8]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    old_names = data_features.column_names(data)
    if len(new_names) != len(old_names):
        raise ValueError('new_names must be same size as data column_count.')
    return {new_name: data[old_name] for new_name, old_name in zip(new_names, old_names)}


def edit_row_items(data: MutableMapping, index: int, items: Mapping) -> MutableMapping:
    """Return a new dict with row index changed to mapping items values.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_row_items(data, 2, {'x': 33, 'y': 88}) -> {'x': [1, 2, 33], 'y': [6, 7, 88]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}

       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_row_items(data, 0, {'x': 55}) -> {'x': [55, 2, 3], 'y': [6, 7, 8]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_row_items_inplace(new_data, index, items)
    return new_data


def edit_row_values(data: MutableMapping, index: int, values: Sequence) -> MutableMapping:
    """Return a new dict with row index changed to values.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_row_values(data, 1, (22, 77)) -> {'x': [1, 22, 3], 'y': [6, 77, 8]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_row_values_inplace(new_data, index, values)
    return new_data


def edit_column(data: MutableMapping, column_name: str, values: MutableSequence) -> MutableMapping:
    """Returns a new dict with values added to data in named column.
       Overrides existing values if column exists,
       Created new column with values if column does not exist.

       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_column(data, 'x', [4, 5, 6]) -> {'x': [4, 5, 6], 'y': [6, 7, 8]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_column_inplace(new_data, column_name, values)
    return new_data


def edit_value(data: MutableMapping, column_name: str, index: int, value: Any) -> MutableMapping:
    """Return a new table with the value in named column changed at row index.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       edit_value(data, 'y', 2, 88) -> {'x': [1, 2, 3], 'y': [6, 7, 88]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_value_inplace(new_data, column_name, index, value)
    return new_data


def drop_row(data: MutableMapping, index: int) -> MutableMapping:
    """Return a new dict with index row removed from data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       drop_row(data, 0) -> {'x': [2, 3], 'y': [7, 8]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    drop_row_inplace(new_data, index)
    return new_data


def drop_label(labels: Union[None, List], index: int) -> Union[None, List]:
    """If labels exists, drop item at index.
    
       Examples:
       labels = [1, 2, 3, 4]
       drop_label(labels, 1) -> [1, 3, 4]
       labels -> [1, 2, 3, 4]

       labels = None
       drop_label(labels, 1) -> None
       labels -> None
    """
    if labels is None: return
    new_labels = copy.copy_list(labels)
    drop_label_inplace(new_labels, index)
    return new_labels


def drop_column(data: MutableMapping, column_name: str) -> MutableMapping:
    """Return a new dict with the named column removed from data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       drop_column(data, 'y') -> {'x': [1, 2, 3]}
       data -> {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    drop_column_inplace(new_data, column_name)
    return new_data


"""
Module used for converting data format to json and json to data format.
"""

from typing import Dict, List, Mapping, Sequence
import json

from tinytim.rows import iterrows
from tinytim.utils import row_dicts_to_data


def data_to_json_list(data: Mapping[str, Sequence]) -> List[Dict]:
    """
    Example:
    data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    data_to_json_list(data) -> [{'x': 1, 'y': 6},
                                {'x': 2, 'y': 7},
                                {'x': 3, 'y': 8}]
    """
    return [row for _, row in iterrows(data)]


def json_list_to_data(l: List[Dict]) -> Dict[str, list]:
    """
    Example:
    json = [{'x': 1, 'y': 6},
            {'x': 2, 'y': 7},
            {'x': 3, 'y': 8}]
    json_list_to_data(json) -> {'x': [1, 2, 3],
                                'y': [6, 7, 8]}
    """
    return row_dicts_to_data(l)


def data_to_json(data: Mapping[str, Sequence]) -> str:
    """
    Example:
    data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    data_to_json(data) -> '''[{"x": 1, "y": 6},
                              {"x": 2, "y": 7},
                              {"x": 3, "y": 8}]'''
    """
    l: List[Dict] = data_to_json_list(data)
    return json.dumps(l)


def json_to_data(j: str) -> Dict[str, list]:
    """
    Example:
    j = '''[{"x": 1, "y": 6},
            {"x": 2, "y": 7},
            {"x": 3, "y": 8}]'''
    json_to_data(j) -> {'x': [1, 2, 3],
                        'y': [6, 7, 8]}
    """
    l = json.loads(j)
    return json_list_to_data(l)


import copy
from typing import List, MutableMapping


def copy_table(data: MutableMapping) -> MutableMapping:
    return copy.deepcopy(data)


def deepcopy_table(data: MutableMapping) -> dict:
    return {col: copy.deepcopy(values) for col, values in data.items()}


def copy_list(values: List) -> List:
    return copy.copy(values)


def deepcopy_list(values: List) -> List:
    return copy.deepcopy(values)


from collections import defaultdict
from typing import Any, Collection, Dict, Generator, Iterable, List, Mapping
from typing import MutableMapping, Optional, Sequence, Tuple


def uniques(values: Iterable) -> List:
    """Return a list of the unique items in values.

       Examples:
       values = [1, 1, 2, 4, 5, 2, 0, 6, 1]
       uniques(values) -> [1, 2, 4, 5, 0, 6]
    """
    out = []
    for value in values:
        if value not in out:
            out.append(value)
    return out


def row_value_tuples(data: MutableMapping, column_names: Collection[str]) -> Tuple[tuple]:
    """Return row value tuples for only columns in column_names.

       data = {'x': [1, 2, 3], 'y': [6, 7, 8], 'z': [9, 10, 11]}
       row_value_tuples(data, ['x', 'z']) -> ((1, 9), (2, 10), (3, 11))
    """
    return tuple(zip(*[data[col] for col in column_names]))


def _keys(key, by) -> dict:
    keys = {}
    if isinstance(by, str):
        keys[by] = key
    else:
        for col, k in zip(by, key):
            keys[col] = k
    return keys


def row_dicts_to_data(rows: Collection[dict], missing_value=None) -> Dict[str, list]:
    """Convert a list of row dicts to dict[col_name: values] format.

       Examples:
       rows = [{'x': 1, 'y': 20}, {'x': 2, 'y': 21}, {'x': 3, 'y': 22}]
       row_dicts_to_data(rows) -> {'x': [1, 2, 3], 'y': [20, 21, 22]}

       rows = [{'x': 1, 'y': 20}, {'x': 2}, {'x': 3, 'y': 22}]
       row_dicts_to_data(rows) -> {'x': [1, 2, 3], 'y': [20, None, 22]}
    """
    keys = all_keys(rows)
    data = defaultdict(list)
    for row in rows:
        for col in keys:
            if col in row:
                data[col].append(row[col])
            else:
                data[col].append(missing_value)
    return dict(data)


def all_bool(values: Collection) -> bool:
    """Return if all items in values are bool or not.
       
       Examples:
       values = [1, True, False, True]
       all_bool(values) -> False

       values = [True, True, False, True]
       all_bool(values) -> True
    """
    return all(isinstance(item, bool) for item in values)


def has_mapping_attrs(obj: Any) -> bool:
    """Check if object has all Mapping attrs.
    
       Examples:
       obj = dict()
       has_mapping_attrs(obj) -> True

       obj = list()
       has_mapping_attrs(obj) -> False
    """
    mapping_attrs = ['__getitem__', '__iter__', '__len__',
                     '__contains__', 'keys', 'items', 'values',
                     'get', '__eq__', '__ne__']
    return all(hasattr(obj, a) for a in mapping_attrs)


def all_keys(dicts: Collection[dict]) -> List:
    """Return all the unique keys from a collection of dicts.
    
       Examples:
       dicts = [{'x': 1, 'y': 2}, {'x': 4, 'z': 7}]
       all_keys(dicts) -> ['x', 'y', 'z']
    """
    keys = []
    for d in dicts:
        for key in d:
            if key not in keys:
                keys.append(key)
    return keys


def row_values_generator(row: Mapping) -> Generator[Any, None, None]:
    """Return a generator that yields values from a row dict.
       
       Examples:
       row = {'x': 1, 'y': 4, 'z': 8}
       generator = row_values_generator(row)
       next(generator) -> 1
       next(generator) -> 4
       next(generator) -> 8
       next(generator) -> StopIteration
    """
    for key in row:
        yield row[key]


def slice_to_range(s: slice, stop: Optional[int] = None) -> range:
    """Convert an int:int:int slice object to a range object.
       Needs stop if s.stop is None since range is not allowed to have stop=None.

       Examples:
       s = slice(1, 4, 0)
       slice_to_range(s) -> ValueError

       s = slice(0, 3, 1)
       slice_to_range(s) -> range(0, 3, 1)
    """
    step = 1 if s.step is None else s.step
    if step == 0:
        raise ValueError('step must not be zero')

    if step > 0:
        start = 0 if s.start is None else s.start
        stop = s.stop if s.stop is not None else stop
    else:
        start = stop if s.start is None else s.start
        if isinstance(start, int):
            start -= 1
        stop = -1 if s.stop is None else s.stop

        if start is None:
            raise ValueError('start cannot be None is range with negative step')

    if stop is None:
        raise ValueError('stop cannot be None in range')
    
    return range(start, stop, step)


def combine_names_rows(column_names: Sequence, rows: Sequence) -> Dict[str, List]:
    """Convert a sequence of column names and a sequence of row values into dict[column_name: values] format.

       Examples:
       column_names = ['x', 'y']
       rows = ((1, 2), (4, 5), (8, 10))
       combine_names_rows(column_names, rows) -> {'x': [1, 4, 8], 'y': [2, 5, 10]}
    """
    return dict(zip(column_names, map(list, zip(*rows))))


def nunique(data: MutableMapping) -> Dict[str, int]:
    """Count number of distinct values in each column.
       Return dict with number of distinct values.

       Examples:
       data = {'x': [1, 2, 2], 'y': [6, 7, 8], 'z': [9, 9, 9]}
       nunique(data) -> {'x': 2, 'y': 3, 'z': 1}
    """
    return {col: len(uniques(values)) for col, values in data.items()}


from typing import Any, Mapping, MutableSequence, Tuple


def column_count(data: Mapping) -> int:
    """Return the number of columns in data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       column_count(data) -> 2
    """
    return len(data)


def row_count(data: Mapping) -> int:
    """Return the number of rows in data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       row_count(data) -> 3
    """
    if column_count(data) == 0: return 0
    return len(data[first_column_name(data)])


def shape(data: Mapping) -> Tuple[int, int]:
    """Return data row count, column count tuple.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       shape(data) -> (3, 2)
    """
    col_count = column_count(data)
    if col_count == 0: return 0, 0
    return row_count(data), col_count


def size(data: Mapping) -> int:
    """Return data row count multiplied by column count.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       size(data) -> 6
    """
    rows, columns = shape(data)
    return rows * columns


def first_column_name(data: Mapping) -> str:
    """Return the name of the first column.
       Raises StopIteration if data has zero columns.

       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       first_column_name(data) -> 'x'
    """
    return next(iter(data))


def column_names(data: Mapping[str, Any]) -> Tuple[str]:
    """Return data column names.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       column_names(data) -> ('x', 'y')
    """
    return tuple(data)


def head(data: Mapping, n: int = 5) -> dict:
    """Return the first n rows of data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       head(data, 2) -> {'x': [1, 2], 'y': [6, 7]}
    """
    return {col: values[:n] for col, values in data.items()}


def tail(data: Mapping, n: int = 5) -> dict:
    """Return the last n rows of data.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       tail(data, 2) -> {'x': [2, 3], 'y': [7, 8]}
    """
    return {col: values[-n:] for col, values in data.items()}


def index(data: Mapping) -> Tuple[int]:
    """Return tuple of data column indexes.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       index(data) -> (0, 1, 2)
    """
    return tuple(range(row_count(data)))


def table_value(data: Mapping, column_name: str, index: int) -> Any:
    """Return one value from column at row index.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       table_value(data, 'x', 1) -> 2
    """
    return data[column_name][index]


def column_values(data: Mapping[str, MutableSequence], column_name: str) -> MutableSequence:
    """Return all the values from one column.
    
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       column_values(data, 'y') -> [6, 7, 8]
    """
    return data[column_name]


from typing import Dict, Generator, Mapping, Sequence, Tuple

import tinytim.data as data_features


def column_dict(data: Mapping[str, Sequence], col: str) -> Dict[str, Sequence]:
    """Return a dict of {col_name, col_values} from data.

       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       column_dict(data, 'x') -> {'x': [1, 2, 3]}
       column_dict(data, 'y') -> {'y': [6, 7, 8]}
    """
    return {col: data[col]}


def itercolumns(data: Mapping) -> Generator[Tuple[str, tuple], None, None]:
    """Return a generator of tuple column name, column values.
       
       Example:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       cols = list(itercolumns(data))
       cols[0] -> ('x', (1, 2, 3)) 
       cols[1] -> ('y', (6, 7, 8))
    """
    for col in data_features.column_names(data):
        yield col, tuple(data[col])


from typing import Generator, Mapping, MutableMapping, Tuple

import tinytim.data as data_features


def row_dict(data: Mapping, index: int) -> dict: 
    """Return one row from data at index.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       row_dict(data, 1) -> {'x': 2, 'y': 7}
    """
    return {col: data_features.table_value(data, col, index) for col in data_features.column_names(data)}


def row_values(data: MutableMapping, index: int) -> tuple:
    """Return a tuple of the values at row index.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       row_values(data, 0) -> (1, 6)
    """
    return tuple(values[index] for values in data.values())


def iterrows(data: Mapping) -> Generator[Tuple[int, dict], None, None]:
    """Return a generator of tuple row index, row dict values.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       generator = iterrows(data)
       next(generator) -> (0, {'x': 1, 'y': 6})
       next(generator) -> (1, {'x': 2, 'y': 7})
       next(generator) -> (2, {'x': 3, 'y': 8})
       next(generator) -> StopIteration
    """
    for i in data_features.index(data):
        yield i, row_dict(data, i)


def itertuples(data: MutableMapping) -> Generator[tuple, None, None]:
    """Return a generator of tuple row values.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       generator = itertuples(data)
       next(generator) -> (1, 6)
       next(generator) -> (2, 7)
       next(generator) -> (3, 8)
       next(generator) -> StopIteration
    """
    for _, row in iterrows(data):
        yield tuple(row.values())


def itervalues(data: MutableMapping) -> Generator[tuple, None, None]:
    """Return a generator of tuple row values.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       generator = itertuples(data)
       next(generator) -> (1, 6)
       next(generator) -> (2, 7)
       next(generator) -> (3, 8)
       next(generator) -> StopIteration
    """
    for _, row in iterrows(data):
        yield tuple(row.values())


def values(data: MutableMapping) -> Tuple[tuple]:
    """Return tuple of tuple row values.
    
       Examples:
       data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
       values(data) -> ((1, 6), (2, 7), (3, 8))
    """
    return tuple(itervalues(data))